#ifndef _HEX_BUFFER_H_
#define _HEX_BUFFER_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif 

#include <iostream.h>

#include <qfile.h>
#include <qfont.h>
#include <qlist.h>
#include <qpainter.h>
#include <qstring.h>

#include "kapp.h"

#define EDIT_INS 0
#define EDIT_OVR 1
#define THIS_FPTR( func ) ((this->*func))

#define UNDO_OK  1
#define REDO_OK  2

#include "hexeditstate.h"

struct SSearchAny
{
  uint operation;
};

struct SSearchGoto
{
  uint operation;
  uint offset;
  uint bit;
  bool backward;
  bool fromCursor;
};

struct SSearchFind
{
  uint operation;
  const char *pattern;
  uint size;
  bool step;
  bool backward;
  bool fromCursor;
  bool inSelection;
};

struct SSearchReplace
{
  uint operation;
  const char *pattern;
  uint pSize;
  const char *replacement;
  uint rSize;
  bool step;
  bool backward;
  bool fromCursor;
  bool inSelection;
  bool prompt;
};

union USearchData
{
  enum EOperation
  {
    opGoto = 0,
    opFind,
    opReplace
  };
  SSearchAny mAny;
  SSearchGoto mGoto;
  SSearchFind mFind;
  SSearchReplace mReplace;
};


struct SInsertData
{
  uint size;
  const char *pattern;
  uint patternSize;
  uint offset;
  bool repeatPattern;
  bool onCursor;
};

struct SExportData
{
  enum EMode
  {
    Everything = 0,
    SelectedArea,  
    Specified   
  }; 

  QString url;
  EMode mode;
  uint start;
  uint stop;
  bool asPostScript;
  bool toStdout;
  bool success;
};



struct SCursorConfig
{
  SCursorConfig( void ) 
  {
    state = 0;
  }

  bool selectOn( void )
  {
    return( state & ShiftButton );
  }

  bool removeSelection( void )
  {
    return( state & ShiftButton ? false : true );
  }

  void setKeepSelection( bool val )
  {
    state = val == true ? state|ShiftButton : state&~ShiftButton;
  }
  
  bool controlButton( void )
  {
    return( state & ControlButton ? true : false );
  }

  bool shiftButton( void )
  {
    return( state & ShiftButton ? true : false );
  }

  bool altButton( void )
  {
    return( state & AltButton ? true : false );
  }


  void emulateControlButton( bool val )
  {
    state = val == true ? state|ControlButton : state&~ControlButton;
  }

  int state;
};



struct SCursorSpec
{
  enum EShape
  {
    solid = 0,
    frame,
    thin,
  };
    
  void reset( void ) 
  { 
    offset = 0;
    x1 = x2 = y = cell = maxCell = 0; 
  }

  void setShape( EShape primaryShape, EShape secondaryShape, uint unitWidth,
		 uint numCell )
  {

    if( primaryShape == thin )
    {
      if( onlyBlock == true ) { primaryShape = solid; }
    }
    if( secondaryShape == thin )
    {
      if( onlyBlock == true ) { secondaryShape = solid; }
    }

    mPrimaryShape = primaryShape;
    if( mPrimaryShape == solid )
    {
      drawWidth1  = unitWidth;
      drawOffset1 = 0;
    }
    else if( mPrimaryShape == frame )
    {
      drawWidth1  = unitWidth * numCell;
      drawOffset1 = 0;
    }
    else
    {
      mPrimaryShape = thin;
      if( thickState == true )
      {
	drawWidth1  = 2;
	drawOffset1 = -1;
      }
      else
      {
	drawWidth1  = 5;
	drawOffset1 = -3;
      }
    }

    mSecondaryShape = secondaryShape;
    if( mSecondaryShape == solid )
    {
      drawWidth2  = unitWidth;
      drawOffset2 = 0;
    }
    else if( mSecondaryShape == frame )
    {
      drawWidth2  = unitWidth * numCell;
      drawOffset2 = 0;
    }
    else
    {
      mSecondaryShape = thin;
      if( thickState == true )
      {
	drawWidth2  = 2;
	drawOffset2 = -1;
      }
      else
      {
	drawWidth2  = 5;
	drawOffset2 = -3;
      }
    }
  }

  void setShapeModifier( bool alwaysBlock, bool useThick )
  {
    onlyBlock  = alwaysBlock;
    thickState = useThick;
  }

  void dump( void )
  {
    cout << "offset: " << offset << " y: " << y << " x1: " << x1 << endl;
    cout << "x2: " << x2 << " cell: " << cell << endl;
  }

  bool inside( uint min, uint max )
  {
    return( offset >= min && offset < max );
  }

  int drawX1( void )
  {
    return( x1 + drawOffset1 );
  }

  int drawX2( void )
  {
    return( x2 + drawOffset2 );
  }

  int width1( void )
  {
    return( drawWidth1 );
  }

  int width2( void )
  {
    return( drawWidth2 );
  }

  uint offset;
  int  y;
  int  x1;
  int  x2;
  int  cell;
  int  maxCell;
  bool onlyBlock;
  bool thickState;

  char data;

  int mPrimaryShape;
  int mSecondaryShape;
  int drawWidth1;
  int drawWidth2;
  int drawOffset1;
  int drawOffset2;
};


struct SCursor
{
  SCursor( void )
  {
    mLineSize = 1;
    mDocumentSize = 0;
    mFixedSizeMode = false;
    mArea = 0;
  }

  void reset( void )
  {
    prev.reset();
    curr.reset();
    next.reset();
  }

  void setDocumentSize( uint documentSize )
  {
    mDocumentSize = documentSize;
  }

  void setFixedSizeMode( bool fixedSizeMode )
  {
    mFixedSizeMode = fixedSizeMode;
  }

  bool incCell( void )
  {
    if( curr.cell + 1 >= curr.maxCell )
    {
      addOffset( 1 );
      return( true );
    }
    else
    {
      next.cell = curr.cell + 1;
      return( false );
    }
  }

  bool decCell( void )
  {
    if( curr.cell == 0 )
    {
      if( curr.offset > 0 )
      {
	decOffset( 1, true );
	next.cell = curr.maxCell > 0 ? curr.maxCell - 1 : 0;
      }
      return( true );
    }
    else
    {
      next.cell = curr.cell - 1;
      return( false );
    }
  }

  void resetCell( void )
  {
    next.cell = 0;
  }


  void addOffset( uint val )
  {
    setOffset( curr.offset + val );
  }

  void decOffset( uint val, bool ignoreCell )
  {
    if( ignoreCell == true || curr.cell == 0 )
    {
      setOffset( val > curr.offset ? curr.offset%mLineSize : curr.offset-val);
    }
    else
    {
      setOffset( curr.offset );
    }
  }

  void setOffset( uint offset )
  {
    /*
    if( offset >= mDocumentSize )
    {
      if( mFixedSizeMode == true )
      {
	offset = mDocumentSize > 0 ? mDocumentSize - 1 : 0;
      }
      else
      {
	offset = mDocumentSize;
      }
    }
    */

    next.offset = offset;
    next.cell = 0;
  }

  uint getOffset( void )
  {
    return( curr.offset );
  }

  void setOffset( uint offset, uint bit, bool backward, bool fromCursor,
		  uint textSize )
  {
    if( fromCursor == true )
    {
      if( backward == true )
      {
	setOffset( offset > curr.offset ? 0 : curr.offset - offset );
      }
      else
      {
	setOffset( curr.offset + offset );
      }
    }
    else
    {
      if( backward == true )
      {
	setOffset( offset > textSize ? 0 : textSize - offset );
      }
      else
      {
	setOffset( offset > textSize ? textSize : offset );
      }
    } 
    setBit( bit );
  }


  void setBit( uint bit )
  {
    bit = (bit > 7) ? 0 : 7 - bit;
    next.cell = bit / mCellWeight;
  }


  void home( bool toExtreme )
  {
    if( toExtreme == true )
    {
      setOffset( 0 );
    }
    else
    {
      setOffset( next.offset - curr.offset % mLineSize );
    }
  }

  void end( bool toExtreme )
  {
    uint maxOffset;
    if( mFixedSizeMode == true )
    {
      maxOffset = mDocumentSize > 0 ? mDocumentSize-1 : mDocumentSize;
    }
    else
    {
      maxOffset = mDocumentSize;
    }

    if( toExtreme == true )
    {
      setOffset( maxOffset );
    }
    else
    {
      uint newOffset = next.offset + (mLineSize-1) - curr.offset % mLineSize;
      setOffset( newOffset > maxOffset ? maxOffset : newOffset );
    }

  }

  void up( uint numLines )
  {
    decOffset( numLines * mLineSize, true );
  }

  void down( uint numLines )
  {
    addOffset( numLines * mLineSize );
  }

  bool changed( void )
  {
    return( prev.offset != curr.offset || prev.cell != curr.cell ||
	    prev.data != curr.data ? true : false );
  }

  uint bit( void )
  {
    uint bitValue = mCellWeight*(curr.maxCell - curr.cell) - 1;
    return( bitValue > 7 ? 7 : bitValue );
  }

  void setLineSize( uint lineSize )
  {
    mLineSize = lineSize < 1 ? 1 : lineSize;
  }

  void setCellWeight( uint cellWeight )
  {
    mCellWeight = cellWeight;
  }

  int area( void )
  {
    return( mArea );
  }

  void setArea( int area )
  {
    mArea = area;
  }

  void setShape( SCursorSpec::EShape primary, SCursorSpec::EShape secondary,
		 uint unitWidth, uint numCell )
  {
    curr.setShape( primary, secondary, unitWidth, numCell );
  }

  void setShapeModifier( bool alwaysBlock, bool useThick )
  {
    curr.setShapeModifier( alwaysBlock, useThick );
  }

  uint mLineSize;
  uint mDocumentSize;
  uint mCellWeight;
  int  mArea;
  bool mFixedSizeMode;
  SCursorSpec prev;
  SCursorSpec curr;
  SCursorSpec next;
};


struct SCursorPosition
{
  int x;
  int w;
};

struct SCursorOffset
{
  uint offset;
  uint bit;
};






struct SHexInputMode
{
  SHexInputMode( void )
  {
    readOnly = false;
    allowResize = true;
  };

  bool readOnly;
  bool allowResize;
};



struct SSelectSpec
{
  void init( uint offset )
  {
    start = stop = anchor = offset;
  }

  void set( uint offset )
  {
    if( offset < anchor )
    {
      start = offset;
      stop  = anchor;
    }
    else
    {
      stop  = offset;
      start = anchor;
    }
  }

  void expand( uint value )
  {
    set( stop + value );
  }

  void shrink( uint value )
  {
    if( start + value > stop )
    {
      set( start );
    }
    else
    {
      set( stop - value );
    }
  }


  uint start;
  uint stop;
  uint anchor;
};


struct SSelect
{
  bool init( uint offset )
  {
    curr.init( offset );
    if( valid == true )
    {
      valid = false;
      return( true );
    }
    else
    {
      return( false );
    }
  }
  
  bool set( uint offset )
  {
    valid = true;

    curr.set( offset );
    if( curr.start != prev.start || curr.stop != prev.stop )
    {
      return( true );
    }
    else
    {
      return( false );
    }
  }

  void expand( uint value )
  {
    if( valid == false )
    {
      return;
    }
    curr.expand( value );
  }

  void shrink( uint value )
  {
    if( valid == false )
    {
      return;
    }
    curr.shrink( value );
  }


  void sync( void )
  { 
    prev = curr;
  }

  bool inside( uint offset )
  {
    return( valid == true && offset >= curr.start && offset < curr.stop );
  }

  bool inside( uint offset, uint range )
  {
    return( valid == true && (offset + range) >= curr.start && 
	    offset < curr.stop );
  }

  bool valid;
  SSelectSpec prev;
  SSelectSpec curr;
};



class CHexAction 
{
  public:
    enum HexAction 
    { 
      replace
    };
    
  public:
    CHexAction( HexAction action, uint offset );
    ~CHexAction( void );
    void setData( uint size, char *data, uint dataSize );

  public:
    HexAction mAction;
    uint mOffset;
    uint mSize;
    char *mData;
    uint mDataSize;
    CHexAction *mNext;
};

class CHexActionGroup 
{
  public:
    CHexActionGroup( uint startOffset, uint startBit );
    ~CHexActionGroup( void );
    void insertAction( CHexAction *hexAction );

  public:
    uint mStartOffset;
    uint mStopOffset;
    uint mStartBit;
    uint mStopBit;
    CHexAction *mHexAction;
};



struct SCursorState
{ 
  bool valid;
  bool selectionValid;
  uint offset;
  uint cell;
  unsigned char data[8];
  uint undoState;
};

struct SFileState
{ 
  bool valid;
  uint size;
  bool modified;
};





class CHexBuffer;
typedef void (CHexBuffer::*PrintCellFunc)( char *buf, unsigned char data );
typedef bool (CHexBuffer::*InputCellFunc)( unsigned char *dest, int value, 
					   uint cell );
typedef void (CHexBuffer::*PrintOffsetFunc)( char *buf, uint offset );


class CHexBuffer : public QByteArray
{
  public:
    enum EEditArea
    {
      edit_none = 0,
      edit_primary,
      edit_secondary,
    };

    enum ECursorMode
    {
      cursor_curr = 0,
      cursor_prev
    };

    enum EEditMode
    {
      edit_insert = 0,
      edit_overwrite,
    };

  public:
    CHexBuffer( void );
    ~CHexBuffer( void );

    int  setLayout( SDisplayLayout &layout );
    void setColor( SDisplayColor &color );

    void setInputMode( SHexInputMode &mode );

    bool matchWidth( uint width );

    
    void setForegroundColor( const QColor &fg1, const QColor &fg2, 
			     const QColor &fg3 );
    void setBackgroundColor( const QColor &bg );
    void setSelectionColor( const QColor &fg, const QColor &bg );
    void setCursorColor( const QColor &actFg, const QColor &actBg,
			 const QColor &pasFg, const QColor &pasBg );

    void setSeparatorColor( const QColor &left, const QColor &right );
    void setOffsetColor( const QColor &color );
    

    void setNonPrintChar( char nonPrintChar );
    void setShowCursor( bool showCursor );
    void setDisableCursor( bool disableCursor );
    void setCursorShapeModifier( bool alwaysBlock, bool thickInsert );
    void setEditMode( EEditMode editMode );
    void setEditMode( EEditMode editMode, bool alwaysBlock, bool thickInsert );
    void setMaximumSize( uint size );
    void setDocumentSize( uint size );
    void setUndoLevel( uint level );
    void setSoundState( bool inputSound, bool fatalSound );
   
    bool loadFile( QFile *file, const QString &url );
    bool insertFile( QFile *file, const QString &url );
    bool saveFile( QFile &file );
    void closeFile( void ); 

    void setFont( const char *family, int pointSize = 12, 
		  int weight = QFont::Normal, bool italic = FALSE );
    void setFont( QFont &font );

    void drawText( QPainter &paint, uint line, int sx, int x, int w );


    int  drawCursor( QPainter &paint, int cursorMode, int area, 
		     bool doDraw, int &x, int &y );
    void drawCursor( QPainter &paint, int startX );

    int  setCursor( int x, int y, bool init );

    bool inputAtCursor( int key );
    bool inputAtCursor( const char *buf, uint bufSize, uint oldSize );
    bool removeAtCursor( bool beforeCursor );
    bool selectedData( QByteArray &array, bool formatted );
    bool exportData( QByteArray &array, uint startLine, uint stopLine );
    bool export( QByteArray &array, SExportData &ed );

    uint printLine( char *dst, uint line );

    bool cutSelection( void );
    bool undo( void );
    bool redo( void );
    int  addBookmark( int position );
    bool removeBookmark( void );

    bool search( USearchData &sd, bool replaceData );
    uint replaceAll( USearchData &sd );

    void doActionGroup( CHexActionGroup *group );
    void doAction( CHexAction *action );

    inline SCursorState &cursorState( void );
    inline SFileState &fileState( void );
    inline const SDisplayLayout &layout( void );

    inline const SHexInputMode &inputMode( void );
    inline const QList<SCursorOffset> &bookmarkList( void );

    inline bool isValid( void );
    inline bool modified( void );
    inline uint undoState( void );
    inline uint documentSize( void );
    inline QString &url( void );
    inline bool hasFileName( void );
    //inline bool onDisk( void );
    inline void setUrl( const QString &url );

    inline bool stepActiveEditArea( void );
    inline int  validEditArea( int activeEditArea );

    inline uint calculateLine( uint offset );
    inline int  lineSize( void );
    inline int  lineHeight( void );
    inline int  fontAscent( void );
    inline int  lineWidth( void );
    inline int  unitWidth( void );
    inline int  numLines( void );
    inline int  totalHeight( void );
    inline const QFont &font( void );
    inline SCursor *textCursor( void );
    inline const QColor &backgroundColor( void );
    inline int  startX( void );
    inline int  startY( void );
    inline void setStartX( int val );
    inline void setStartY( int val );

    inline bool selectionSet( uint offset, bool init );
    inline uint startSelection( void );
    inline uint stopSelection( void );
    inline uint firstSelection( void );
    inline uint lastSelection( void );
    inline void selectionSyncronize( void );
    inline void selectionStartChange( uint &offset1, uint &offset2 );
    inline void selectionStopChange( uint &offset1, uint &offset2 );
    inline bool selectionValid( void );

    inline uint cursorOffset( void );
    inline uint cursorBit( void );
    inline uint cursorLine( void );
    inline uint prevCursorLine( void );
    inline SCursor &cursor( void );

    inline void currCursor( EEditArea editArea, SCursorPosition &p );
    inline void prevCursor( EEditArea editArea, SCursorPosition &p );

    inline void cursorUp( uint lines );
    inline void cursorDown( uint lines );
    inline void cursorRight( bool cellLevel );
    inline void cursorLeft( bool cellLevel );
    inline void cursorStep( uint size, bool forward, bool modulo );
    inline void cursorHome( bool toExtreme );
    inline void cursorEnd( bool toExtreme );
    inline void cursorGoto( uint offset, uint bit );
    inline void cursorGoto( uint offset, uint bit, bool backward, 
			    bool fromCursor );
    inline bool cursorChanged( void );
    inline void cursorResetEditArea( void );
    inline bool cursorPrimaryEdit( void );
    inline int  cursorFixedPosition( int position, int height );
    inline int  cursorChangePosition( int position, int height );
    void cursorReset( void );
    

  private:
    void computeLineWidth( void );
    void computeNumLines( void );
    void cursorCompute( void );


    void recordStart( SCursor &cursor );
    void recordReplace( SCursor &cursor, uint size, char *data, uint dataSize);
    void recordEnd( SCursor &cursor );
    void doReplace( CHexAction *hexAction, bool removeData );


    inline const QColor &foregroundColor( uint column );
    inline bool insideSelection( uint offset );
    inline bool insideSelection( uint offset, uint range );
    inline uint startSelection( uint offset );
    inline uint stopSelection( uint offset );

    inline void printDummyCell( char *buf, unsigned char data );
    inline void printHexadecimalBigCell( char *buf, unsigned char data );
    inline void printHexadecimalSmallCell( char *buf, unsigned char data );
    inline void printOctalCell( char *buf, unsigned char data );
    inline void printBinaryCell( char *buf, unsigned char data );
    inline void printAsciiCell( char *buf, unsigned char data );
    inline void printDummyOffset( char *buf, uint offset );
    inline void printHexadecimalBigOffset( char *buf, uint offset );
    inline void printHexadecimalSmallOffset( char *buf, uint offset );
    inline void printDecimalOffset( char *buf, uint offset );

    bool inputDummy( unsigned char *dest, int value, uint cell );
    bool inputHexadecimal( unsigned char *dest, int value, uint cell );
    bool inputOctal( unsigned char *dest, int value, uint cell );
    bool inputBinary( unsigned char *dest, int value, uint cell );
    bool inputAscii( unsigned char *dest, int value, uint cell );

    bool moveBuffer( uint destOffset, uint srcOffset );
    bool resizeBuffer( uint offset );

    void inputSound( void );
    void fatalSound( void );

  signals:
    void fileSize( uint size );

  private:
    QFont mFont;
    QString mUrl;

    /*
    QColor mTextPrimaryFg[2];
    QColor mTextSecondaryFg;
    QColor mTextOffsetFg;
    QColor mTextBg;
    QColor mLeftSeparatorFg;
    QColor mRightSeparatorFg;
    QColor mSelectBg;
    QColor mSelectFg;
    QColor mCursorActFg;
    QColor mCursorActBg;
    QColor mCursorPasFg;
    QColor mCursorPasBg;
    */

    char mNonPrintChar;    
    SDisplayLayout mLayout;
    SDisplayColor mColor;

    char *mColorIndex;
    char *mPrintBuf;
    int  mStartX;
    int  mStartY;

    int  mFontHeight;
    int  mFontAscent;
    int  mLineWidth;
    int  mFixedWidth;
    int  mUnitWidth;
    int  mSplitWidth;
    int  mNumLines;
    int  mTextStart1;
    int  mTextStart2;
    int  mNumCell;

    uint mDocumentSize;
    uint mMaximumSize;
    bool mFixedSizeMode;
    bool mDocumentModified;
    EEditMode mEditMode;
    SHexInputMode mInputMode;

    int  mOffsetSize;
    int  mOffsetIndex;
    int  mPrimaryWidth;
    int  mSecondaryWidth;
    int  mActiveEditArea;

    SSelect mSelect;
    SCursor mCursor;
    bool mShowCursor;
    bool mDisableCursor;

    bool mInputErrorSound;
    bool mFatalErrorSound;

    uint mUndoLimit;
    uint mUndoIndex;
    QList<CHexActionGroup> mUndoList;
    QList<SCursorOffset> mBookmarkList;

    PrintCellFunc   printCell;
    PrintOffsetFunc printOffset;
    InputCellFunc   inputCell;

};



inline SCursorState &CHexBuffer::cursorState( void )
{
  static SCursorState state;
  if( size() == 0 )
  {
    state.valid          = false;
    state.selectionValid = false;
    state.offset         = 0;
    state.cell           = 0;
    state.undoState      = 0;
    memset( state.data, 0, sizeof(state.data) );
  }
  else
  {
    state.valid          = true;
    state.selectionValid = selectionValid();
    state.offset         = cursorOffset();
    state.cell           = cursorBit();
    state.undoState      = undoState();

    uint offset = state.offset;
    for( uint i = 0; i < sizeof( state.data ); i++ )
    {
      state.data[i] = (offset+i >= mDocumentSize) ? 0 : data()[offset+i];
    }
    
  }
  return( state );
}


inline const SDisplayLayout &CHexBuffer::layout( void )
{
  return( mLayout );
}



inline const SHexInputMode &CHexBuffer::inputMode( void )
{
  return( mInputMode );
}

inline const QList<SCursorOffset> &CHexBuffer::bookmarkList( void )
{
  return( mBookmarkList );
}




inline SFileState &CHexBuffer::fileState( void )
{
  static SFileState state;
  if( size() == 0 )
  {
    state.valid    = false;
    state.size     = 0;
    state.modified = false;

  }
  else
  {
    state.valid    = true;
    state.size     = mDocumentSize;
    state.modified = mDocumentModified;
  }

  return( state );
}








inline bool CHexBuffer::modified( void )
{
  return( mDocumentModified );
}

inline uint CHexBuffer::undoState( void )
{
  return( (mUndoIndex > 0 ? UNDO_OK : 0) | 
	  (mUndoIndex < mUndoList.count() ? REDO_OK : 0) );
}


inline uint CHexBuffer::documentSize( void )
{
  return( mDocumentSize );
}






inline QString &CHexBuffer::url( void )
{
  return( mUrl );
}


inline bool CHexBuffer::isValid( void )
{
  return( size() == 0 ? false : true );
}


inline bool CHexBuffer::hasFileName( void )
{
  if( mUrl.isEmpty() || mUrl.contains( i18n( "Untitled" ), false ) )
  {
    return( false );
  }
  else
  {
    return( true );
  }
}


inline void CHexBuffer::setUrl( const QString &url )
{
  mUrl = url.data(); // deep copy
}




inline bool CHexBuffer::stepActiveEditArea( void )
{
  bool changed;
  if( mLayout.secondaryMode == SDisplayLayout::hide )
  {
    changed = mActiveEditArea == edit_secondary ? true : false;
    mActiveEditArea = edit_primary;
  }
  else
  {
    changed = true;
    mActiveEditArea = mActiveEditArea == edit_primary ? 
      edit_secondary : edit_primary;
  }

  setEditMode( mEditMode ); // Sets the cursor shapes as well

  if( changed == true )
  {
    mCursor.resetCell();
    cursorCompute();
  }
  
  return( changed );
}


inline int CHexBuffer::validEditArea( int activeEditArea )
{
  if( mLayout.secondaryMode == SDisplayLayout::hide )
  {
    return( edit_primary );
  }
  else
  {
    return( activeEditArea );
  }
}






inline uint CHexBuffer::calculateLine( uint offset )
{
  return( mLayout.lineSize == 0 ? 0 : offset / mLayout.lineSize );
}


inline bool CHexBuffer::selectionSet( uint offset, bool init )
{
  if( offset >= size() ) { offset = size() > 0 ? size() - 1 : 0; }

  if( init == true )
  {
    return( mSelect.init( offset ) );
  }
  else
  {
    return( mSelect.set( offset ) );
  }
}


inline const QColor &CHexBuffer::foregroundColor( uint column )  
{
  if( column > mLayout.lineSize )
  {
    return( black );
  }
  else
  {
    return( mColor.primaryFg[ mColorIndex[ column ] ] );
  }
}


inline bool CHexBuffer::insideSelection( uint offset )
{
  return( mSelect.inside( offset ) );
}


inline bool CHexBuffer::insideSelection( uint offset, uint range )
{
  return( mSelect.inside( offset, range ) );
}


inline uint CHexBuffer::startSelection( uint offset )
{
  return( mSelect.curr.start < offset ? 0 : mSelect.curr.start - offset );
}

inline uint CHexBuffer::stopSelection( uint offset )
{
  return( mSelect.curr.stop > offset + mLayout.lineSize ? 
	  mLayout.lineSize : mSelect.curr.stop - offset );
}


inline uint CHexBuffer::startSelection( void )
{
  return( mSelect.curr.start );
}

inline uint CHexBuffer::stopSelection( void )
{
  return( mSelect.curr.stop );
}

inline uint CHexBuffer::firstSelection( void )
{
  return( mSelect.curr.start < mSelect.prev.start ? mSelect.curr.start : 
	  mSelect.prev.start );
}

inline uint CHexBuffer::lastSelection( void )
{
  return( mSelect.curr.stop > mSelect.prev.stop ? mSelect.curr.stop : 
	  mSelect.prev.stop );
}


inline void CHexBuffer::selectionSyncronize( void )
{
  mSelect.sync();
}


inline void CHexBuffer::selectionStartChange( uint &offset1, uint &offset2 )
{
  if( mSelect.curr.start == mSelect.prev.start )
  {
    offset1 = offset2 = mSelect.curr.start;
  }
  else if( mSelect.curr.start < mSelect.prev.start )
  { 
    offset1 = mSelect.curr.start;
    offset2 = mSelect.prev.start;
  }
  else
  {
    offset1 = mSelect.prev.start;
    offset2 = mSelect.curr.start;
  }
}



inline void CHexBuffer::selectionStopChange( uint &offset1, uint &offset2 )
{
  if( mSelect.curr.stop == mSelect.prev.stop )
  {
    offset1 = offset2 = mSelect.curr.stop;
  }
  else if( mSelect.curr.stop < mSelect.prev.stop )
  { 
    offset1 = mSelect.curr.stop;
    offset2 = mSelect.prev.stop;
  }
  else
  {
    offset1 = mSelect.prev.stop;
    offset2 = mSelect.curr.stop;
  }
}


inline bool CHexBuffer::selectionValid( void )
{
  return( mSelect.valid );
}



inline uint CHexBuffer::cursorOffset( void )
{
  return( mCursor.curr.offset );
}

inline uint CHexBuffer::cursorBit( void )
{
  return( mCursor.bit() );
}


inline uint CHexBuffer::cursorLine( void )
{
  return( mCursor.curr.y / lineHeight() );
}

inline uint CHexBuffer::prevCursorLine( void )
{
  return( mCursor.prev.y / lineHeight() );
}

inline SCursor &CHexBuffer::cursor( void )
{
  return( mCursor );
}


inline void CHexBuffer::currCursor( EEditArea editArea, SCursorPosition &p )
{
  if( editArea == edit_primary )
  {
    if( mActiveEditArea == edit_primary )
    {
      p.x = mCursor.curr.drawX1();
      p.w = mCursor.curr.width1();
    }
    else
    {
      p.x = mCursor.curr.drawX2();
      p.w = mUnitWidth;
    }
  }
  else
  {
    if( mActiveEditArea == edit_primary )
    {
      p.x = mCursor.curr.drawX2();
      p.w = mUnitWidth;
    }
    else
    {
      p.x = mCursor.curr.drawX1();
      p.w = mUnitWidth * mNumCell;
    }
  }




  /*
  if( editArea == mActiveEditArea )
  {
    p.x = mCursor.curr.drawX1();
    p.w = 12;//mCursor.curr.width1();
  }
  else
  {
    p.x = mCursor.curr.drawX2();
    p.w = 12;//mCursor.curr.width2();
  }
  */
}

inline void CHexBuffer::prevCursor( EEditArea editArea, SCursorPosition &p )
{

  if( editArea == edit_primary )
  {
    if( mActiveEditArea == edit_primary )
    {
      p.x = mCursor.prev.drawX1();
      p.w = mUnitWidth * mNumCell;
    }
    else
    {
      p.x = mCursor.prev.drawX2();
      p.w = mUnitWidth;
    }
  }
  else
  {
    if( mActiveEditArea == edit_primary )
    {
      p.x = mCursor.prev.drawX2();
      p.w = mUnitWidth;
    }
    else
    {
      p.x = mCursor.prev.drawX1();
      p.w = mUnitWidth * mNumCell;
    }
  }


  /*
  if( editArea == edit_primary )
  {
    if( mActiveEditArea == edit_primary )
    {
      p.x = mCursor.prev.drawX1();
      p.w = mCursor.prev.width1();
    }
    else
    {
      p.x = mCursor.prev.drawX1();
      p.w = mUnitWidth * mNumCell;
    }
  }
  else
  {
    if( mActiveEditArea == edit_primary )
    {
      p.x = mCursor.curr.drawX2();
      p.w = mUnitWidth;
    }
    else
    {
      p.x = mCursor.curr.drawX2();
      p.w = mCursor.curr.width2();
    }
  }
  */

  /*
  if( editArea == mActiveEditArea )
  {
    p.x = mCursor.prev.drawX1();
    p.w = 12;//mCursor.prev.width1();
  }
  else
  {
    p.x = mCursor.prev.drawX2();
    p.w = 12;//mCursor.prev.width2();
  }
  */
}






inline void CHexBuffer::cursorUp( uint lines )
{
  mCursor.up( lines );
  cursorCompute();
}

inline void CHexBuffer::cursorDown( uint lines )
{
  mCursor.down( lines );
  cursorCompute();
}


inline void CHexBuffer::cursorRight( bool cellLevel )
{
  if( cellLevel == true && mActiveEditArea == edit_primary )
  {
    mCursor.incCell();
  }
  else
  {
    mCursor.addOffset( 1 );
  }
  cursorCompute();
}


inline void CHexBuffer::cursorStep( uint size, bool forward, bool modulo )
{
  if( forward == true )
  {
    if( modulo == true )
    {
      uint offset = mCursor.getOffset() + size;
      mCursor.setOffset( offset - offset % size );
    }
    else
    {
      mCursor.addOffset( size );
    }
  }
  else
  {
    if( modulo == true )
    {
      uint offset = mCursor.getOffset();
      if( offset % size )
      {
	mCursor.decOffset( offset % size, false );
      }
      else
      {
	mCursor.setOffset( offset < size ? 0 : offset - size );
      }
    }
    else
    {
      mCursor.decOffset( size, false );
    }
  }
  cursorCompute();
}


inline void CHexBuffer::cursorLeft( bool cellLevel )
{
  if( cellLevel == true && mActiveEditArea == edit_primary )
  {
    mCursor.decCell();
  }
  else
  {
    mCursor.decOffset( 1, false );
  }
  cursorCompute();
}

inline void CHexBuffer::cursorHome( bool toExtreme )
{
  mCursor.home( toExtreme );
  cursorCompute();
}

inline void CHexBuffer::cursorEnd( bool toExtreme )
{
  mCursor.end( toExtreme );
  cursorCompute();
}


inline void CHexBuffer::cursorGoto( uint offset, uint bit, bool backward, 
				    bool fromCursor )
{
  uint maxOffset = mFixedSizeMode == true ? mMaximumSize-1 : documentSize();
  mCursor.setOffset( offset, bit, backward, fromCursor, maxOffset );
  cursorCompute();
}



inline void CHexBuffer::cursorGoto( uint offset, uint bit )
{
  mCursor.setOffset( offset );
  mCursor.setBit( bit );
  cursorCompute();
}



inline bool CHexBuffer::cursorChanged( void )
{
  return( mCursor.changed() );
}

inline void CHexBuffer::cursorResetEditArea( void )
{
  mCursor.setArea( edit_none );
}

inline bool CHexBuffer::cursorPrimaryEdit( void )
{
  return( mActiveEditArea == edit_primary ? true : false );
}


inline int CHexBuffer::cursorFixedPosition( int position, int height )
{
  position += mCursor.curr.y - mCursor.prev.y;
  if( position < 0 )
  {
    return( 0 );
  }
  else if( position + height > totalHeight() )
  {
    return( height > totalHeight() ? 0 : totalHeight() - height );
  }
  else
  {
    if( mCursor.curr.y < position )
    {
      return( mCursor.curr.y );
    }
    else if( mCursor.curr.y > position + height )
    {
      return( mCursor.curr.y - height + lineHeight() );
    }
    else
    {
      return( position );
    }
  }
}


inline int CHexBuffer::cursorChangePosition( int position, int height )
{
  if( mCursor.curr.y < position || mCursor.curr.y > position + height )
  {
    // When cursor is at top of window
    position = mCursor.curr.y;
  }
  else if( mCursor.curr.y > position + height - lineHeight() )
  {
    // When cursor is at bottom of window
    position = mCursor.curr.y - height + lineHeight();
  }

  return( position );
}


inline void CHexBuffer::printDummyCell( char *buf, unsigned char data )
{
  buf[0] = 0;
}

inline void CHexBuffer::printHexadecimalBigCell( char *buf, 
						 unsigned char data )
{
  static char hexBuffer[16]=
  {
    '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
  };
  buf[0] = hexBuffer[ (data>>4)&0x0F ];
  buf[1] = hexBuffer[ data&0x0F ];
}


inline void CHexBuffer::printHexadecimalSmallCell( char *buf, 
						   unsigned char data )
{
  static char hexBuffer[16]=
  {
    '0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'
  };
  buf[0] = hexBuffer[ (data>>4)&0x0F ];
  buf[1] = hexBuffer[ data&0x0F ];
}





inline void CHexBuffer::printOctalCell( char *buf, unsigned char data )
{
  static char hexBuffer[8]=
  {
    '0','1','2','3','4','5','6','7'
  };

  buf[0] = hexBuffer[ (data>>6)&0x07 ];
  buf[1] = hexBuffer[ (data>>3)&0x07 ];
  buf[2] = hexBuffer[ data&0x07 ];
}

inline void CHexBuffer::printBinaryCell( char *buf, unsigned char data )
{
  for( int i = 0; i < 8; i++ )
  {
    buf[7-i] = (data&(1<<i)) ? '1' : '0';
  }
}

inline void CHexBuffer::printAsciiCell( char *buf, unsigned char data )
{
  buf[0] = data < 32 ? mNonPrintChar : data;
}


inline void CHexBuffer::printDummyOffset( char *buf, uint offset )
{
  buf[0] = 0;
}

inline void CHexBuffer::printHexadecimalBigOffset( char *buf, uint offset )
{
  sprintf( buf, "%04X:%04X", offset>>16, offset&0x0000FFFF );
}

inline void CHexBuffer::printHexadecimalSmallOffset( char *buf, uint offset )
{
  sprintf( buf, "%04x:%04x", offset>>16, offset&0x0000FFFF );
}

inline void CHexBuffer::printDecimalOffset( char *buf, uint offset )
{
  sprintf( buf, "%010u", offset );
}


inline int CHexBuffer::lineSize( void )
{
  return( mLayout.lineSize );
}


inline int CHexBuffer::lineHeight( void )
{
  return( mFontHeight );
}

inline int CHexBuffer::fontAscent( void )
{
  return( mFontAscent );
}

inline int CHexBuffer::lineWidth( void )
{
  return( mLineWidth );
}

inline int CHexBuffer::unitWidth( void )
{
  return( mUnitWidth );
}

inline int CHexBuffer::numLines( void )
{
  return( mNumLines );
}

inline int CHexBuffer::totalHeight( void )
{
  return( mNumLines * mFontHeight );
}

inline const QFont &CHexBuffer::font( void )
{
  return( mFont );
}

inline SCursor *CHexBuffer::textCursor( void )
{
  return( &mCursor );
}


inline const QColor &CHexBuffer::backgroundColor( void )
{
  return( isValid() == true ? mColor.textBg : mColor.inactiveBg );
}


inline int CHexBuffer::startX( void )
{
  return( mStartX );
} 

inline int CHexBuffer::startY( void )
{
  return( mStartY );
}

inline void CHexBuffer::setStartX( int val )
{
  mStartX = val;
}

inline void CHexBuffer::setStartY( int val )
{
  mStartY = val;
}







#endif

